//$(
//=======================================================================
//      File: $RCSfile: rt_watchdog.c,v $
//      Version: $Revision: $  Dated: $Date:  $
//
//=======================================================================
//
//
//  FILE: rt_watchdog.c
//
//  DESCRIPTION:
//
//	Thread to initialize and update the watchdog timer on a periodic
//	basis. It uses the PC's CMOS Real-Time Clock (RTC). The RTC
//	may be programed to cause an alarm interrupt (IRQ8) if its alarm
//	time is reached. This thread keeps pushing the alarm time
//	forward so that timeout doesn't happen (under normal operation).
//	If the thread is not allowed to run before the timeout time
//	expires, an IRQ8 interrupt occurs. A jumper on some embedded PC
//	boards (e.g., jumper W24 on the AMPRO Little Board /P5e) may be
//	set to cause a board RESET when this happens. The system will then
//	re-boot and restart.
//	
//
//=======================================================================
//$)
//
#include <watchdog.h>

/* All globals defined in main program file */
extern int thread_ready;

static int fd;
static struct rtc_time rtc_tm;

static void rtWatchdogCleanup( void * );


//
// Thread to start and manage a watchdog alarm.
//
void *rtWatchdogThread( void *arg )
{
    /* Perform thread initialization operations here */

    /* Initialize watchdog timer */
    DPRINTF("Initializing watchdog timer.\n");
    rtWatchdogInit();

    /* Set cleanup handler to close RTC interface */
    pthread_cleanup_push( rtWatchdogCleanup, (void *)NULL );

    /* After initialization, set ready flag */
    thread_ready = 1;

    /* Wait for "go" condition */
    wait_for_go( THREAD_1 );

    while( 1 ) {

	/* Sleep for an alarm-sleep time */
	DPRINTF("Sleeping...\n");
	sleep( WATCHDOG_ALARM_SLEEP_SECS );

        /* Advance the watchdog timer by an alarm-advance time */
	DPRINTF("\n\nAdvancing watchdog timer.\n");
	rtWatchdogAdvance( WATCHDOG_ALARM_ADVANCE_SECS );

    }

    sched_yield();

    /* Must match pthread_cleanup_push() with pthread_cleanup_pop()! */
    pthread_cleanup_pop( 1 );
}



//
// Routing to initialize the watchdog timer and set its first
// timeout alarm time.
//
int rtWatchdogInit ( void )
{
    int retval;

    /* Open the Real-Time Clock (RTC) device */
    fd = open ("/dev/rtc", O_RDONLY);

    if (fd ==  -1) {
	errno_abort( "/dev/rtc");
    }

    /* Read the RTC time/date */
    retval = ioctl(fd, RTC_RD_TIME, &rtc_tm);
    if (retval == -1) {
	errno_abort( "ioctl: RTC_RD_TIME error" );
    }

#if defined DEBUG

fprintf(stderr, "Current RTC date/time is %d-%d-%d, %02d:%02d:%02d.\n",
	rtc_tm.tm_mday, rtc_tm.tm_mon + 1, rtc_tm.tm_year + 1900,
	rtc_tm.tm_hour, rtc_tm.tm_min, rtc_tm.tm_sec);

#endif
   /*
    * Set the alarm to WATCHDOG_ALARM_INIT_SECS in the future,
    * and check for rollover.
    */
    rtc_tm.tm_sec += WATCHDOG_ALARM_INIT_SECS;

    while( rtc_tm.tm_sec > 59 ) {
	rtc_tm.tm_sec -= 60;
	rtc_tm.tm_min++;
    }

    while( rtc_tm.tm_min > 59 ) {
	rtc_tm.tm_min -= 60;
	rtc_tm.tm_hour++;
    }

    while( rtc_tm.tm_hour > 23 )
	rtc_tm.tm_hour -= 24;

    retval = ioctl(fd, RTC_ALM_SET, &rtc_tm);

    if (retval == -1) {
	errno_abort( "ioctl: RTC_ALM_SET error" );
    }

#if defined DEBUG

fprintf(stderr, "Alarm time now set to %02d:%02d:%02d.\n",
	rtc_tm.tm_hour, rtc_tm.tm_min, rtc_tm.tm_sec);

#endif

    /* Disable Update interrupts (must do this!) */
    retval = ioctl(fd, RTC_UIE_OFF, 0);
    if (retval == -1) {
	errno_abort( "ioctl: RTC_UIE_OFF error" );
    }

    /* Disable Periodic interrupts (not necessary, but safe) */
    retval = ioctl(fd, RTC_PIE_OFF, 0);
    if (retval == -1) {
	errno_abort( "ioctl: RTC_PIE_OFF error" );
    }

    /* Enable Alarm interrupts */
    retval = ioctl(fd, RTC_AIE_ON, 0);
    if (retval == -1) {
	errno_abort( "ioctl: RTC_AIE_ON error" );
    }
}



//
// Routine to advance the watchdog timeout alarm time to
// 'alarmSecs'.
//
void rtWatchdogAdvance( int alarmSecs )
{
    int retval;

    /* Read the RTC time/date */
    retval = ioctl(fd, RTC_RD_TIME, &rtc_tm);
    if (retval == -1) {
	errno_abort( "ioctl: RTC_RD_TIME error" );
    }

#if defined DEBUG

fprintf(stderr, "Current RTC date/time is %d-%d-%d, %02d:%02d:%02d.\n",
	rtc_tm.tm_mday, rtc_tm.tm_mon + 1, rtc_tm.tm_year + 1900,
	rtc_tm.tm_hour, rtc_tm.tm_min, rtc_tm.tm_sec);

#endif

   /*
    * Set the alarm to 'alarmSecs' in the future,
    * and check for rollover.
    */
    rtc_tm.tm_sec += alarmSecs;

    while( rtc_tm.tm_sec > 59 ) {
	rtc_tm.tm_sec -= 60;
	rtc_tm.tm_min++;
    }

    while( rtc_tm.tm_min > 59 ) {
	rtc_tm.tm_min -= 60;
	rtc_tm.tm_hour++;
    }

    while( rtc_tm.tm_hour > 23 )
	rtc_tm.tm_hour -= 24;

    retval = ioctl(fd, RTC_ALM_SET, &rtc_tm);

    if (retval == -1) {
	errno_abort( "ioctl: RTC_ALM_SET error" );
    }

#if defined DEBUG

fprintf(stderr, "Alarm time now set to %02d:%02d:%02d.\n",
	rtc_tm.tm_hour, rtc_tm.tm_min, rtc_tm.tm_sec);

#endif

}



//
// Exit cleanup handler for watchdog thread.
//
static void rtWatchdogCleanup( void *val )
{
    int retval;

    /* Disable alarm interrupts */
    retval = ioctl(fd, RTC_AIE_OFF, 0);

    if (retval == -1) {
	errno_abort( "ioctl: RTC_AIE_OFF error" );

    }

    close(fd);
}


